@@ -8,7 +8,8 @@ import pandas as pd
 from Bio.Phylo import NewickIO, write
 from ete3 import Tree
 
-from pastml import col_name2cat, value2list, STATES, METHOD, CHARACTER, get_personalized_feature_name, NUM_SCENARIOS
+from pastml import col_name2cat, value2list, STATES, METHOD, CHARACTER, get_personalized_feature_name, NUM_SCENARIOS, \
+    numeric2datetime, datetime2numeric
 from pastml.annotation import preannotate_forest, get_forest_stats
 from pastml.file import get_combined_ancestral_state_file, get_named_tree_file, get_pastml_parameter_file, \
     get_pastml_marginal_prob_file, get_pastml_work_dir
@@ -19,35 +20,21 @@ from pastml.models.hky import KAPPA, HKY_STATES, HKY
 from pastml.models.jtt import JTT_STATES, JTT
 from pastml.parsimony import is_parsimonious, parsimonious_acr, ACCTRAN, DELTRAN, DOWNPASS, MP_METHODS, MP, \
     get_default_mp_method
-from pastml.tree import name_tree, collapse_zero_branches, annotate_dates, DATE, read_forest
-from pastml.visualisation.cytoscape_manager import visualize, TIMELINE_SAMPLED, TIMELINE_NODES, TIMELINE_LTT
+from pastml.tree import name_tree, collapse_zero_branches, annotate_dates, DATE, read_forest, DATE_CI
+from pastml.visualisation.cytoscape_manager import visualize, TIMELINE_SAMPLED, TIMELINE_NODES, TIMELINE_LTT, \
+    DIST_TO_ROOT_LABEL, DATE_LABEL
 from pastml.visualisation.itol_manager import generate_itol_annotations
 from pastml.visualisation.tree_compressor import REASONABLE_NUMBER_OF_TIPS
+from pastml.visualisation import get_formatted_date
 
-PASTML_VERSION = '1.9.25'
+PASTML_VERSION = '1.9.27'
 
 warnings.filterwarnings("ignore", append=True)
 
 COPY = 'COPY'
 
 
-def datetime2numeric(d):
-    """
-    Converts a datetime date to numeric format.
-    For example: 2016-12-31 -> 2016.9972677595629; 2016-1-1 -> 2016.0
-    :param d: a date to be converted
-    :type d: np.datetime
-    :return: numeric representation of the date
-    :rtype: float
-    """
-    first_jan_this_year = pd.datetime(year=d.year, month=1, day=1)
-    day_of_this_year = d - first_jan_this_year
-    first_jan_next_year = pd.datetime(year=d.year + 1, month=1, day=1)
-    days_in_this_year = first_jan_next_year - first_jan_this_year
-    return d.year + day_of_this_year / days_in_this_year
-
-
-def _parse_pastml_parameters(params, states):
+def _parse_pastml_parameters(params, states, reoptimise=False):
     logger = logging.getLogger('pastml')
     frequencies, sf, kappa = None, None, None
     if not isinstance(params, str) and not isinstance(params, dict):
@@ -74,24 +61,28 @@ def _parse_pastml_parameters(params, states):
     params = {str(k.encode('ASCII', 'replace').decode()): v for (k, v) in params.items()}
     frequencies_specified = set(states) & set(params.keys())
     if frequencies_specified:
-        if len(frequencies_specified) < len(states):
+        if len(frequencies_specified) < len(states) and not reoptimise:
             logger.error('Some frequency parameters are specified, but missing the following states: {}, '
                          'ignoring specified frequencies.'.format(', '.join(set(states) - frequencies_specified)))
         else:
-            frequencies = np.array([params[state] for state in states])
             try:
+                min_freq = min(float(params[state]) for state in frequencies_specified if float(params[state]) > 0) / 10
+                frequencies = np.array([params[state] if state in params.keys() else min_freq for state in states])
                 frequencies = frequencies.astype(np.float64)
-                if np.round(frequencies.sum() - 1, 3) != 0:
+                if np.round(frequencies.sum() - 1, 3) != 0 and not reoptimise:
                     logger.error('Specified frequencies ({}) do not sum up to one,'
                                  'ignoring them.'.format(frequencies))
                     frequencies = None
                 elif np.any(frequencies < 0):
-                    logger.error('Some of the specified frequencies ({}) are negative,'
-                                 'ignoring them.'.format(frequencies))
-                    frequencies = None
+                    if not reoptimise:
+                        logger.error('Some of the specified frequencies ({}) are negative,'
+                                     'ignoring them.'.format(frequencies))
+                        frequencies = None
+                    else:
+                        frequencies = np.maximum(frequencies, min_freq)
                 frequencies /= frequencies.sum()
             except:
-                logger.error('Specified frequencies ({}) must not be negative, ignoring them.'.format(frequencies))
+                logger.error('Specified frequencies ({}) must be float, ignoring them.'.format(frequencies))
                 frequencies = None
     if SCALING_FACTOR in params:
         sf = params[SCALING_FACTOR]
@@ -153,7 +144,7 @@ def _serialize_acr(args):
 
 
 def reconstruct_ancestral_states(forest, character, states, prediction_method=MPPA, model=F81,
-                                 params=None, avg_br_len=None, num_nodes=None, num_tips=None,
+                                 params=None, reoptimise=False, avg_br_len=None, num_nodes=None, num_tips=None,
                                  force_joint=True):
     """
     Reconstructs ancestral states for the given character on the given tree.
@@ -184,6 +175,9 @@ def reconstruct_ancestral_states(forest, character, states, prediction_method=MP
         e.g. set to 1 to keep the original branches). Could also be in a form path_to_param_file.
         Only makes sense for ML methods.
     :type params: dict or str
+    :param reoptimise: (False by default) if set to True and the parameters are specified,
+        they will be considered as an optimisation starting point instead, and the parameters will be optimised.
+    :type reoptimise: bool
 
     :return: ACR result dictionary whose values depend on the prediction method.
     :rtype: dict
@@ -211,7 +205,7 @@ def reconstruct_ancestral_states(forest, character, states, prediction_method=MP
         return ml_acr(forest=forest, character=character, prediction_method=prediction_method, model=model,
                       states=states,
                       avg_br_len=avg_br_len, num_nodes=num_nodes, num_tips=num_tips, freqs=freqs, sf=sf, kappa=kappa,
-                      force_joint=force_joint)
+                      force_joint=force_joint, reoptimise=reoptimise)
     if is_parsimonious(prediction_method):
         return parsimonious_acr(forest, character, prediction_method, states, num_nodes, num_tips)
 
@@ -219,7 +213,8 @@ def reconstruct_ancestral_states(forest, character, states, prediction_method=MP
                      .format(prediction_method, ', '.join(ML_METHODS), ', '.join(MP_METHODS), COPY))
 
 
-def acr(forest, df, prediction_method=MPPA, model=F81, column2parameters=None, force_joint=True, threads=0):
+def acr(forest, df, prediction_method=MPPA, model=F81, column2parameters=None, force_joint=True, threads=0,
+        reoptimise=False):
     """
     Reconstructs ancestral states for the given tree and
     all the characters specified as columns of the given annotation dataframe.
@@ -243,6 +238,9 @@ def acr(forest, df, prediction_method=MPPA, model=F81, column2parameters=None, f
         or pastml.ml.SCALING_FACTOR (then the value should be the scaling factor for three branches,
         e.g. set to 1 to keep the original branches). Could also be in a form {column: path_to_param_file}.
     :type column2parameters: dict
+    :param reoptimise: (False by default) if set to True and the parameters are specified,
+        they will be considered as an optimisation starting point instead, and the parameters will be optimised.
+    :type reoptimise: bool
     :param force_joint: (optional, default is True) whether the JOINT state should be added to the MPPA prediction
         even when not selected by the Brier score
     :type force_joint: bool
@@ -299,11 +297,12 @@ def acr(forest, df, prediction_method=MPPA, model=F81, column2parameters=None, f
         with ThreadPool(processes=threads - 1) as pool:
             acr_results = \
                 pool.map(func=_work, iterable=((forest, column, get_states(method, model, column), method, model,
-                                                column2parameters[column] if column in column2parameters else None)
+                                                (column2parameters[column] if column in column2parameters else None),
+                                                reoptimise)
                                                for (column, method, model) in zip(columns, prediction_methods, models)))
     else:
         acr_results = [_work((forest, column, get_states(method, model, column), method, model,
-                              column2parameters[column] if column in column2parameters else None))
+                              (column2parameters[column] if column in column2parameters else None), reoptimise))
                        for (column, method, model) in zip(columns, prediction_methods, models)]
 
     result = []
@@ -326,7 +325,7 @@ def pastml_pipeline(tree, data, data_sep='\t', id_index=0,
                     tip_size_threshold=REASONABLE_NUMBER_OF_TIPS,
                     out_data=None, html_compressed=None, html=None, work_dir=None,
                     verbose=False, forced_joint=False, upload_to_itol=False, itol_id=None, itol_project=None,
-                    itol_tree_name=None, offline=False, threads=0):
+                    itol_tree_name=None, offline=False, threads=0, reoptimise=False):
     """
     Applies PastML to the given tree(s) with the specified states and visualises the result (as html maps).
 
@@ -385,6 +384,9 @@ def pastml_pipeline(tree, data, data_sep='\t', id_index=0,
         and parameter value - the float frequency value, between 0 and 1),
         and tree branch scaling factor (parameter name pastml.ml.SCALING_FACTOR).
     :type parameters: str or list(str) or dict
+    :param reoptimise: (False by default) if set to True and the parameters are specified,
+        they will be considered as an optimisation starting point instead, and optimised.
+    :type reoptimise: bool
 
     :param name_column: (optional) name of the annotation table column to be used for node names
         in the compressed map visualisation
@@ -450,13 +452,15 @@ def pastml_pipeline(tree, data, data_sep='\t', id_index=0,
     """
     logger = _set_up_pastml_logger(verbose)
 
-    age_label = 'Dist. to root' if root_date is None else 'Date'
-
     roots, df, name_column, root_dates = \
         _validate_input(columns, data, data_sep, root_date if html_compressed or html or upload_to_itol else None,
                         id_index, name_column if html_compressed else None, tree,
                         copy_only=COPY == prediction_method or (isinstance(prediction_method, list)
                                                                 and all(COPY == _ for _ in prediction_method)))
+
+    age_label = DIST_TO_ROOT_LABEL \
+        if (root_date is None and not next((True for root in roots if getattr(root, DATE, None) is not None), False)) \
+        else DATE_LABEL
     annotate_dates(roots, root_dates=root_dates)
 
     if parameters:
@@ -479,13 +483,14 @@ def pastml_pipeline(tree, data, data_sep='\t', id_index=0,
         threads = max(os.cpu_count(), 1)
 
     acr_results = acr(roots, df, prediction_method=prediction_method, model=model, column2parameters=parameters,
-                      force_joint=forced_joint, threads=threads)
+                      force_joint=forced_joint, threads=threads, reoptimise=reoptimise)
     column2states = {acr_result[CHARACTER]: acr_result[STATES] for acr_result in acr_results}
 
     if not out_data:
         out_data = os.path.join(work_dir, get_combined_ancestral_state_file())
 
-    state_df = _serialize_predicted_states(sorted(column2states.keys()), out_data, roots)
+    state_df = _serialize_predicted_states(sorted(column2states.keys()), out_data, roots,
+                                           dates_are_dates=age_label==DATE_LABEL)
 
     # a meta-method would have added a suffix to the name feature
     if html_compressed and name_column and name_column not in column2states:
@@ -496,7 +501,7 @@ def pastml_pipeline(tree, data, data_sep='\t', id_index=0,
     itol_result = None
     new_tree = os.path.join(work_dir, get_named_tree_file(tree))
     features = list(column2states.keys())
-    nwks = [root.write(format_root_node=True, format=3, features=[DATE] + features) for root in roots]
+    nwks = [root.write(format_root_node=True, format=3, features=[DATE, DATE_CI] + features) for root in roots]
     with open(new_tree, 'w+') as f:
         f.write('\n'.join(nwks))
     try:
@@ -507,14 +512,6 @@ def pastml_pipeline(tree, data, data_sep='\t', id_index=0,
         with open(nexus, 'r') as f:
             nexus_str = f.read().replace('&&NHX:', '&')
             for feature in features:
-                # # fixing multiple locations: going from A|B|C to {A, B, C}
-                # any_state = '|'.join(column2states[feature])
-                # for multistate in re.findall('[:]{}[=](?:{})(?:[|](?:{}))+'
-                # .format(feature, any_state, any_state), nexus_str):
-                #     nexus_str = nexus_str.replace(multistate,
-                #                                   ':{}={{{}}}'.format(feature,
-                #                                                         multistate.replace(':{}='.format(feature), '')
-                #                                                         .replace('|', ',')))
                 nexus_str = nexus_str.replace(':{}='.format(feature), ',{}='.format(feature))
         with open(nexus, 'w') as f:
             f.write(nexus_str)
@@ -527,6 +524,12 @@ def pastml_pipeline(tree, data, data_sep='\t', id_index=0,
         pool = ThreadPool(processes=threads - 1)
         async_result = pool.map_async(func=_serialize_acr, iterable=((acr_res, work_dir) for acr_res in acr_results))
         if upload_to_itol:
+            if DATE_LABEL == age_label:
+                try:
+                    dates = state_df[DATE].apply(lambda _: numeric2datetime(_).strftime("%d %b %Y"))
+                    state_df[DATE] = dates
+                except:
+                    pass
             itol_result = pool.apply_async(func=generate_itol_annotations,
                                            args=(column2states, work_dir, acr_results, state_df, age_label,
                                                  new_tree, itol_id, itol_project, itol_tree_name))
@@ -534,6 +537,12 @@ def pastml_pipeline(tree, data, data_sep='\t', id_index=0,
         for acr_res in acr_results:
             _serialize_acr((acr_res, work_dir))
         if upload_to_itol:
+            if DATE_LABEL == age_label:
+                try:
+                    dates = state_df[DATE].apply(lambda _: numeric2datetime(_).strftime("%d %b %Y"))
+                    state_df[DATE] = dates
+                except:
+                    pass
             generate_itol_annotations(column2states, work_dir, acr_results, state_df, age_label,
                                       new_tree, itol_id, itol_project, itol_tree_name)
 
@@ -645,14 +654,14 @@ def _validate_input(columns, data, data_sep, root_dates, id_index, name_column,
     return roots, df, name_column, root_dates
 
 
-def _serialize_predicted_states(columns, out_data, roots):
+def _serialize_predicted_states(columns, out_data, roots, dates_are_dates=True):
     ids, data = [], []
     # Not using DataFrames to speed up document writing
     with open(out_data, 'w+') as f:
         f.write('node\t{}\n'.format('\t'.join(columns)))
         for root in roots:
             for node in root.traverse():
-                vs = [node.dist, getattr(node, DATE)]
+                vs = [node.dist, get_formatted_date(node, dates_are_dates)]
                 column2values = {}
                 for column in columns:
                     value = getattr(node, column, set())
@@ -767,6 +776,9 @@ def main():
                                 '(parameter name should be the corresponding state, '
                                 'and parameter value - the float frequency value, between 0 and 1),'
                                 'and tree branch scaling factor (parameter name {}).'.format(SCALING_FACTOR))
+    acr_group.add_argument('--reoptimise', action='store_true',
+                           help='if the parameters are specified, they will be considered as an optimisation '
+                                'starting point instead and optimised.')
 
     vis_group = parser.add_argument_group('visualisation-related arguments')
     vis_group.add_argument('-n', '--name_column', type=str, default=None,
