@@ -1,5 +1,6 @@
 import logging
 import os
+from collections import defaultdict
 from glob import glob
 from queue import Queue
 from shutil import copyfile
@@ -7,10 +8,12 @@ from shutil import copyfile
 import numpy as np
 from jinja2 import Environment, PackageLoader
 
-from pastml.tree import DATE
+from pastml import numeric2datetime
+from pastml.tree import DATE, DATE_CI
+from pastml.visualisation import get_formatted_date
 from pastml.visualisation.colour_generator import get_enough_colours, WHITE
 from pastml.visualisation.tree_compressor import NUM_TIPS_INSIDE, TIPS_INSIDE, TIPS_BELOW, \
-    REASONABLE_NUMBER_OF_TIPS, compress_tree, INTERNAL_NODES_INSIDE, ROOTS, IS_TIP
+    REASONABLE_NUMBER_OF_TIPS, compress_tree, INTERNAL_NODES_INSIDE, ROOTS, IS_TIP, ROOT_DATES
 
 JS_LIST = ["https://pastml.pasteur.fr/static/js/jquery.min.js",
            "https://pastml.pasteur.fr/static/js/jquery.qtip.min.js",
@@ -53,6 +56,10 @@ FONT_SIZE = 'node_fontsize'
 MILESTONE = 'mile'
 
 
+DATE_LABEL = 'date'
+
+DIST_TO_ROOT_LABEL = 'dist. to root'
+
 def get_fake_node(n_id, x, y):
     attributes = {ID: n_id, 'fake': 1}
     return _get_node(attributes, position=(x, y))
@@ -89,7 +96,7 @@ def get_scaling_function(y_m, y_M, x_m, x_M):
 
 
 def set_cyto_features_compressed(n, size_scaling, e_size_scaling, font_scaling, transform_size, transform_e_size, state,
-                                 root_names, suffix=''):
+                                 root_names, root_dates, suffix=''):
     tips_inside, internal_nodes_inside, roots = getattr(n, TIPS_INSIDE, []), \
                                                 getattr(n, INTERNAL_NODES_INSIDE, []), \
                                                 getattr(n, ROOTS, [])
@@ -113,7 +120,10 @@ def set_cyto_features_compressed(n, size_scaling, e_size_scaling, font_scaling,
     n.add_feature('node_{}{}'.format(TIPS_INSIDE, suffix), tips_inside_str)
     n.add_feature('node_{}{}'.format(INTERNAL_NODES_INSIDE, suffix), internal_ns_inside_str)
     n.add_feature('node_{}{}'.format(TIPS_BELOW, suffix), tips_below_str)
-    n.add_feature('node_{}{}'.format(ROOTS, suffix), ', '.join(sorted(root_names)))
+    root_name2date = dict(zip(root_names, root_dates))
+    root_names = sorted(root_names)
+    n.add_feature('node_{}{}'.format(ROOTS, suffix), ', '.join(root_names))
+    n.add_feature('node_{}{}'.format(ROOT_DATES, suffix), ', '.join(str(root_name2date[_]) for _ in root_names))
 
     edge_size = len(roots)
     if edge_size > 1:
@@ -128,7 +138,8 @@ def set_cyto_features_tree(n, state):
     n.add_feature(EDGE_NAME, n.dist)
 
 
-def _forest2json_compressed(forest, compressed_forest, columns, name_feature, get_date, milestones=None):
+def _forest2json_compressed(forest, compressed_forest, columns, name_feature, get_date, milestones=None,
+                            dates_are_dates=True):
     e_size_scaling, font_scaling, size_scaling, transform_e_size, transform_size = \
         get_size_transformations(compressed_forest)
 
@@ -153,14 +164,17 @@ def _forest2json_compressed(forest, compressed_forest, columns, name_feature, ge
                 todo.put_nowait(c)
 
     n2state = {}
+
+
     # Set the cytoscape features
     for compressed_tree in compressed_forest:
         for n in compressed_tree.traverse():
             state = get_column_value_str(n, name_feature, format_list=False, list_value='') if name_feature else ''
             n2state[n] = state
             root_names = [_.name for _ in getattr(n, ROOTS)]
+            root_dates = [get_formatted_date(_, dates_are_dates) for _ in getattr(n, ROOTS)]
             set_cyto_features_compressed(n, size_scaling, e_size_scaling, font_scaling,
-                                         transform_size, transform_e_size, state, root_names)
+                                         transform_size, transform_e_size, state, root_names, root_dates)
 
     # Calculate node coordinates
     min_size = 2 * min(min(getattr(_, NODE_SIZE) for _ in compressed_tree.traverse())
@@ -228,8 +242,15 @@ def _forest2json_compressed(forest, compressed_forest, columns, name_feature, ge
                     if roots_i:
                         n.add_feature(MILESTONE, i)
                         root_names = [getattr(_, BRANCH_NAME) if getattr(_, DATE) > milestone else _.name for _ in roots_i]
+                        root_dates = [getattr(_, DATE) for _ in roots_i]
+                        if dates_are_dates:
+                            try:
+                                root_dates = [numeric2datetime(_).strftime("%d %b %Y") for _ in root_dates]
+                            except:
+                                pass
                         set_cyto_features_compressed(n, size_scaling, e_size_scaling, font_scaling, transform_size,
-                                                     transform_e_size, state, root_names=root_names, suffix=suffix)
+                                                     transform_e_size, state, root_names=root_names, suffix=suffix,
+                                                     root_dates=root_dates)
                         nodes_i.append(n)
                 nodes = nodes_i
 
@@ -261,7 +282,8 @@ def _forest2json_compressed(forest, compressed_forest, columns, name_feature, ge
     return json_dict, sorted(clazzes)
 
 
-def _forest2json(forest, columns, name_feature, get_date, milestones=None, timeline_type=TIMELINE_SAMPLED):
+def _forest2json(forest, columns, name_feature, get_date, milestones=None, timeline_type=TIMELINE_SAMPLED,
+                 dates_are_dates=True):
 
     min_root_date = min(getattr(tree, DATE) for tree in forest)
     width = sum(len(tree) for tree in forest)
@@ -281,6 +303,7 @@ def _forest2json(forest, columns, name_feature, get_date, milestones=None, timel
         for n in tree.traverse('postorder'):
             state = get_column_value_str(n, name_feature, format_list=False, list_value='') if name_feature else ''
             n.add_feature('node_root_id', n.name)
+            n.add_feature('node_root_date', get_formatted_date(n, dates_are_dates))
             if not n.is_leaf():
                 n2x[n] = np.mean([n2x[_] for _ in n.children])
             n2y[n] = (getattr(n, DATE) - min_root_date) * height_factor
@@ -393,7 +416,8 @@ def get_tooltip(n, columns):
 
 
 def save_as_cytoscape_html(forest, out_html, column2states, name_feature, name2colour, compressed_forest,
-                           milestone_label, timeline_type, milestones, get_date, work_dir, local_css_js=False):
+                           milestone_label, timeline_type, milestones, get_date, work_dir, local_css_js=False,
+                           milestone_labels=None):
     """
     Converts a forest to an html representation using Cytoscape.js.
 
@@ -409,15 +433,17 @@ def save_as_cytoscape_html(forest, out_html, column2states, name_feature, name2c
     """
     graph_name = os.path.splitext(os.path.basename(out_html))[0]
     columns = sorted(column2states.keys())
+    if milestone_labels is None:
+        milestone_labels = ['{:g}'.format(_) for _ in milestones]
 
     if compressed_forest is not None:
         json_dict, clazzes \
             = _forest2json_compressed(forest, compressed_forest, columns, name_feature=name_feature, get_date=get_date,
-                                      milestones=milestones)
+                                      milestones=milestones, dates_are_dates=milestone_label == DATE_LABEL)
     else:
         json_dict, clazzes \
             = _forest2json(forest, columns, name_feature=name_feature, get_date=get_date, milestones=milestones,
-                           timeline_type=timeline_type)
+                           timeline_type=timeline_type, dates_are_dates=milestone_label == DATE_LABEL)
     loader = PackageLoader('pastml')
     env = Environment(loader=loader)
     template = env.get_template('pie_tree.js') if compressed_forest is not None \
@@ -434,11 +460,12 @@ def save_as_cytoscape_html(forest, out_html, column2states, name_feature, name2c
             """.format(i=i, percent=round(100 / n, 2), colour=name2colour[cat])
         clazz2css[_clazz_list2css_class(clazz_list)] = css
     graph = template.render(clazz2css=clazz2css.items(), elements=json_dict, title=graph_name,
-                            years=['{:g}'.format(_) for _ in milestones],
+                            years=milestone_labels,
                             tips='samples' if TIMELINE_SAMPLED == timeline_type
                             else ('lineages ending' if TIMELINE_LTT == timeline_type else 'external nodes'),
                             internal_nodes='internal nodes' if TIMELINE_NODES == timeline_type
-                            else 'diversification events')
+                            else 'diversification events',
+                            age_label=milestone_label)
     slider = env.get_template('time_slider.html').render(min_date=0, max_date=len(milestones) - 1,
                                                          name=milestone_label) if len(milestones) > 1 else ''
 
@@ -503,10 +530,10 @@ def get_column_value_str(n, column, format_list=True, list_value=''):
 def visualize(forest, column2states, work_dir, name_column=None, html=None, html_compressed=None,
               tip_size_threshold=REASONABLE_NUMBER_OF_TIPS, date_label='Dist. to root', timeline_type=TIMELINE_SAMPLED,
               local_css_js=False):
-
     one_column = next(iter(column2states.keys())) if len(column2states) == 1 else None
 
     name2colour = {}
+    state2color = None
     for column, states in column2states.items():
         num_unique_values = len(states)
         colours = get_enough_colours(num_unique_values)
@@ -556,6 +583,14 @@ def visualize(forest, column2states, work_dir, name_column=None, html=None, html
     milestones = sorted({dates[0], dates[len(dates) // 8], dates[len(dates) // 4], dates[3 * len(dates) // 8],
                          dates[len(dates) // 2], dates[5 * len(dates) // 8], dates[3 * len(dates) // 4],
                          dates[7 * len(dates) // 8], dates[-1]})
+    milestone_labels = None
+    if DATE_LABEL == date_label:
+        try:
+            milestone_labels = [numeric2datetime(_).strftime("%d %b %Y") for _ in milestones]
+        except:
+            pass
+    if milestone_labels is None:
+        milestone_labels = ['{:g}'.format(_) for _ in milestones]
 
     if html:
         total_num_tips = sum(len(tree) for tree in forest)
@@ -569,9 +604,10 @@ def visualize(forest, column2states, work_dir, name_column=None, html=None, html
             save_as_cytoscape_html(forest, html, column2states=column2states, name2colour=name2colour,
                                    name_feature='name', compressed_forest=None, milestone_label=date_label,
                                    timeline_type=timeline_type, milestones=milestones, get_date=get_date,
-                                   work_dir=work_dir, local_css_js=local_css_js)
+                                   work_dir=work_dir, local_css_js=local_css_js, milestone_labels=milestone_labels)
 
     if html_compressed:
+        resolve_polytomies(column2states, forest, name_column, state2color)
         compressed_forest = [compress_tree(tree, columns=column2states.keys(), tip_size_threshold=tip_size_threshold)
                              for tree in forest]
 
@@ -582,13 +618,69 @@ def visualize(forest, column2states, work_dir, name_column=None, html=None, html
             for _ in (tree.traverse() if timeline_type in [TIMELINE_LTT, TIMELINE_NODES] else tree):
                 first_date = min(first_date, getattr(_, DATE))
                 last_date = max(last_date, getattr(_, DATE))
+
         milestones = [ms for ms in milestones if first_date <= ms <= last_date]
         if milestones[0] > first_date:
             milestones.insert(0, first_date)
         if milestones[-1] < last_date:
             milestones.append(last_date)
 
+        if DATE_LABEL == date_label:
+            try:
+                milestone_labels = [numeric2datetime(_).strftime("%d %b %Y") for _ in milestones]
+            except:
+                pass
+        if milestone_labels is None:
+            milestone_labels = ['{:g}'.format(_) for _ in milestones]
+
         save_as_cytoscape_html(forest, html_compressed, column2states=column2states, name2colour=name2colour,
                                name_feature=name_column, compressed_forest=compressed_forest,
                                milestone_label=date_label, timeline_type=timeline_type,
-                               milestones=milestones, get_date=get_date, work_dir=work_dir, local_css_js=local_css_js)
+                               milestones=milestones, get_date=get_date, work_dir=work_dir, local_css_js=local_css_js,
+                               milestone_labels=milestone_labels)
+
+
+def resolve_polytomies(column2states, forest, name_column, state2color):
+    columns = sorted(column2states.keys())
+
+    def get_prediction(n):
+        return tuple(tuple(sorted(getattr(n, c, set()))) for c in columns)
+
+    for tree in forest:
+        todo = [tree]
+        while todo:
+            n = todo.pop()
+            todo.extend(n.children)
+            if len(n.children) > 2:
+                n_date = getattr(n, DATE)
+                n_ci = getattr(n, DATE_CI, None)
+                state2children = defaultdict(list)
+                for c in n.children:
+                    state2children[get_prediction(c)].append(c)
+                if len(state2children) > 1:
+                    i = 0
+                    for children in state2children.values():
+                        if len(children) > 1:
+                            pol = n.add_child(dist=0, name='{}.polytomy_{}'.format(n.name, i))
+                            i += 1
+                            child = children[0]
+                            pol.add_feature(DATE, n_date)
+                            pol.add_feature(DATE_CI, n_ci)
+                            for c in columns:
+                                pol.add_feature(c, getattr(child, c))
+                            if hasattr(child, UNRESOLVED):
+                                pol.add_feature(UNRESOLVED, getattr(child, UNRESOLVED))
+                            if name_column is not None:
+                                sts = getattr(pol, name_column, set())
+                                if len(sts) == 1:
+                                    col = state2color[next(iter(sts))]
+                                    if getattr(n, name_column, set()) == sts:
+                                        pol.add_feature('edge_color', col)
+                                    for c in children:
+                                        c.add_feature('edge_color', col)
+                            pol.add_feature(BRANCH_NAME, '{}-{}'.format(n.name, pol.name))
+                            for c in children:
+                                c.add_feature(BRANCH_NAME, '{}-{}'.format(pol.name, c.name))
+                            for c in children:
+                                n.remove_child(c)
+                                pol.add_child(c)
